/*
 * Run programs residing in the filesystem.
 *
 * Copyright (C) 2022, 2023, 2024 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/sys/err.h>

#include <fsclient/client.h>
#include <fsclient/process.h>
#include <systypes/fcntl.h>
#include <systypes/format.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ops.h"



/* Transfer size for communication. */

static const offset_t TO_TRANSFER = 1024;



/* Initiated programs (jobs).
   NOTE: Arbitrary limit. */

#define NUMBER_OF_JOBS 32

static file_t *readers[NUMBER_OF_JOBS] = {NULL};
static process_t *processes[NUMBER_OF_JOBS] = {NULL};
static const char *programs[NUMBER_OF_JOBS] = {NULL};
static int next_job = 0;



/* Wait for a program to finish, showing its output. */

static int _wait_program(file_t *reader, process_t *process)
{
  char buffer[TO_TRANSFER];
  notify_flags_t flags;
  notify_values_t values;
  long err;

  /* Read until the pipe yields no more data. */

  while (1)
  {
    offset_t nread = client_read(reader, buffer, TO_TRANSFER);

    if (nread)
      fwrite(buffer, sizeof(char), nread, stdout);
    else
      break;
  }

  /* Close the pipe and obtain the process state. */

  client_close(reader);

  err = process_wait(process, &flags, &values);

  printf("End process (flags %" pFMTnotify_flags "x values: %ld, %ld)\n", flags, values.sig, values.val);

  if (err)
    return -1;

  return values.val;
}

/* Run the given program. */

int run_program(int argc, char *argv[])
{
  process_t *process;
  file_t *reader, *writer;
  int last_job;
  long err;

  /* Create a pipe for process output. */

  err = client_pipe(&reader, &writer, 0);

  if (err)
  {
    printf("Could not obtain pipe: %s\n", l4sys_errtostr(err));
    return -1;
  }

  /* Start the process. */

  err = process_spawn(argc, (const char **) argv, writer, &process);

  if (err)
  {
    printf("Could not start process: %s\n", l4sys_errtostr(err));
    return -1;
  }

  printf("Finished program initiation.\n");

  /* Release the writing end of the pipe. */

  client_close(writer);

  /* Record the output stream. */

  last_job = next_job;

  while (readers[next_job] != NULL)
  {
    next_job++;

    if (next_job >= NUMBER_OF_JOBS)
      next_job = 0;

    /* Wait for the process to complete if no more job slots are available. */

    if (next_job == last_job)
      return _wait_program(reader, process);
  }

  readers[next_job] = reader;
  processes[next_job] = process;
  programs[next_job] = strdup(argv[0]);

  return 0;
}

/* Show initiated programs. */

int show_programs(int argc, char *argv[])
{
  (void) argc; (void) argv;

  int job_number;

  for (job_number = 0; job_number < NUMBER_OF_JOBS; job_number++)
  {
    if (readers[job_number] != NULL)
      printf("[%d] %s\n", job_number, programs[job_number]);
  }

  return 0;
}

/* Wait for the indicated program to finish, showing its output. */

int wait_program(int argc, char *argv[])
{
  int job_number;
  int exitcode;

  if (argc < 1)
    return -1;

  job_number = atoi(argv[0]);

  if (readers[job_number] == NULL)
    return -1;

  exitcode = _wait_program(readers[job_number], processes[job_number]);

  readers[job_number] = NULL;
  processes[job_number] = NULL;

  free(programs[job_number]);
  programs[job_number] = NULL;

  return exitcode;
}

/* vim: tabstop=2 expandtab shiftwidth=2
*/
